@@ -6,8 +6,9 @@ import numpy as np
 import matplotlib.pyplot as mpl
 from matplotlib.transforms  import Bbox
 import warnings
-from .tools import kwargs_update
+from .tools import kwargs_update, is_arraylike
 from .decorators import make_method
+
 # remark: do no import .plot.*** at this level
 #         otherwise import loop with astrobject
 __all__ = ["specplot","skyplot","figout"]
@@ -16,7 +17,83 @@ __all__ = ["specplot","skyplot","figout"]
 # =  Axes Add-on           = #
 # ========================== #
 # --------------------- #
-# - Spectrum          - #
+# - Matrix           - #
+# --------------------- #
+@make_method(mpl.Axes)
+def corrmatrix(ax, matrix, npoints=None,
+               cmap=None, significance_range=[0,6], ncolors=6,
+               matrixlabels=None,
+               **kwargs):
+    """
+    In Dev
+    """
+    # --------------
+    # Init Settings
+    
+    shape = np.shape(matrix)
+    def get_significance(rho, npoints):
+        """ """
+        rho = np.abs(rho)
+        return rho * np.sqrt( (npoints-2.) /(1.-rho**2) )
+
+    prop = dict(interpolation="nearest", origin="upper",
+                alpha=0.9)
+    
+    if npoints is None:
+        color_matrix = matrix
+        cmap = mpl.cm.Spectral if cmap is None else \
+          mpl.cm.get_cmap(cmap) if type(cmap) == str else cmap
+
+    else:
+        color_matrix = get_significance(matrix, npoints)
+        cmap = mpl.cm.get_cmap("YlGnBu_r",ncolors) if cmap is None else cmap
+        prop["vmin"],prop["vmax"] = significance_range
+        cticks = np.linspace(prop["vmin"],prop["vmax"],ncolors+1)
+
+
+    # - imshow
+    propimshow = kwargs_update(prop,**kwargs)
+    
+    pl = mpl.imshow(color_matrix, cmap=cmap,**propimshow)
+
+    # - ColorBar
+    cb = mpl.colorbar(pl, ticks=cticks if npoints is not None else None)
+    if npoints is not None:
+        cb.ax.set_yticklabels([r'$%d$'%v if i<(len(cticks)-1) else r'$>%d$'%v for i,v in enumerate(cticks)],
+                                fontsize="large")
+        cb.set_label(r"$\mathrm{Corr.\ Significance\ [in\,\sigma]}$", fontsize="x-large")
+
+    # --------
+    # - Ticks
+    ax.xaxis.set_ticks(np.linspace(0,shape[1]-1,shape[1]))
+    ax.yaxis.set_ticks(np.linspace(0,shape[0]-1,shape[0]))
+
+
+    
+    # - Text
+    
+    _ = [[ax.text(i,j,"%.2f"%matrix[i][j],
+            va="center", ha="center",
+            fontproperties="black",color="k")
+            for i in range(shape[0])] for j in range(shape[1])]
+        
+    if matrixlabels is not None:
+        _ = ax.set_xticklabels(matrixlabels,
+                    fontsize="xx-large")
+        _ = ax.set_yticklabels(matrixlabels,
+                                fontsize="xx-large")
+
+    # - No ticks 
+    [ax_.tick_params(length=0) for ax_ in [ax,cb.ax]]
+    return pl
+
+    
+    
+
+
+
+# --------------------- #
+# - Scatter           - #
 # --------------------- #
 @make_method(mpl.Axes)
 def errorscatter(ax,x,y,dx=None,dy=None,**kwargs):
@@ -284,7 +361,7 @@ def wcsplot(ax, wcs, exp_order=10,
 # --------------------- #
 @make_method(mpl.Axes)
 def voronoi_patchs(ax, xy, c=None, vmax=None, vmin=None,
-                   cmap=mpl.cm.jet,
+                   cmap=None,
                    cbar=True,cblabel="",cbarprop={},
                    **kwargs):
     """
@@ -303,7 +380,7 @@ def voronoi_patchs(ax, xy, c=None, vmax=None, vmin=None,
                                    set.
     - other options -
     
-    cmap: [mpl.cm]                 a colormap
+    cmap: [mpl.cm]                 a colormap. Default would be viridis if possible YlGnBu otherwise
     
     cbar: [bool or ax]             provide here an ax where the colorbar should be
                                    drawn. You can also set True to have a default one
@@ -318,7 +395,12 @@ def voronoi_patchs(ax, xy, c=None, vmax=None, vmin=None,
     """
     from scipy.spatial import Voronoi
     from matplotlib.collections import PolyCollection
-    
+    if cmap is None:
+        try:
+            cmap = mpl.cm.viridis
+        except:
+            warning.warns("Viridis is not available, you should update matplotlib")
+            cmap = mpl.cm.YlGnBu
     # --------------------
     # - Define the Cells
     npoint = np.shape(xy)[0]
@@ -330,7 +412,7 @@ def voronoi_patchs(ax, xy, c=None, vmax=None, vmin=None,
     # --------------------
     # - Color of the Cells
     if c is not None:
-        if "__iter__" not in dir(c):
+        if not is_arraylike(c):
             c = [c]*npoint
         else:
             c = np.asarray(c)
@@ -368,22 +450,41 @@ def voronoi_patchs(ax, xy, c=None, vmax=None, vmin=None,
     # ----------------- #
     if not (color is "None" or not cbar):
         # - this means it is not an ax
-        axcar = ax.insert_ax("right", shrunk=0.8,space=.0,axspace=0.03) \
+        axcar = ax.insert_ax("right", shrunk=0.9,space=.0,axspace=0.03) \
           if "imshow" not in dir(cbar) else cbar
         calpha = cbarprop.pop('alpha',kwargs.pop("alpha",None))
         return collec, axcar.colorbar(cmap,vmin=vmin,vmax=vmax,label=cblabel,
                                       alpha=calpha,**cbarprop)
     # - no cbar
     return collec
-    
+
 # --------------------------- #
 # -    Color Bar            - #
 # --------------------------- #
 @make_method(mpl.Axes)
 def colorbar(ax,cmap,vmin=0,vmax=1,label="",
-             fontsize="x-large",npoint=256,
-            **kwargs):
-    """
+             fontsize="x-large",**kwargs):
+    """ Set a colorbar in the given axis
+
+    Parameters
+    -----------
+    ax: [mpl's Axes]
+        Axis in which the colorbar will be drawn
+
+    cmap: [mpl's colormap]
+        A matplotlib colormap
+
+    vmin, vmax: [float,float] -optional-
+        Extend of the colormap, values of the upper and lower colors
+
+    label, fontsize: [string, string/float] -optional-
+        Label of the colorbar and its associated size
+     
+    **kwargs goes to matplotlib.colobar.ColorbarBase
+
+    Return
+    ------
+    colorbar
     """
     import matplotlib
     if "orientation" not in kwargs.keys():
@@ -566,6 +667,7 @@ def vline(ax,value,ymin=None,ymax=None,
      ax.plot([value,value],[ymin,ymax],
              scalex=False,scaley=False,
              **kwargs)
+     
 @make_method(mpl.Axes)      
 def hline(ax,value,xmin=None,xmax=None,
                 **kwargs):
@@ -602,7 +704,7 @@ def vspan(ax,minvalue,maxvalue,ymin=None,ymax=None,
 # ========================== #
 @make_method(mpl.Figure)
 def figout(fig,savefile=None,show=True,add_thumbnails=False,
-           dpi=200):
+           dpi=200, close=True):
     """This methods parse the show/savefile to know if the figure
     shall the shown or saved."""
     
@@ -611,15 +713,24 @@ def figout(fig,savefile=None,show=True,add_thumbnails=False,
         savefile = None
 
     if savefile is not None:
-        fig.savefig(savefile+'.png',dpi=dpi)
-        fig.savefig(savefile+'.pdf')
+        if not savefile.endswith(".pdf"):
+            extention = ".png" if not savefile.endswith(".png") else ""
+            fig.savefig(savefile+extention,dpi=dpi)
+            
+        if not savefile.endswith(".png"):
+            extention = ".pdf" if not savefile.endswith(".pdf") else ""
+            fig.savefig(savefile+extention)
+            
         if add_thumbnails:
             fig.savefig(savefile+"_thumb"+'.png',dpi=dpi/10.)
             
+        if close: mpl.close(fig)
+            
     elif show:
         fig.canvas.draw()
         fig.show()
         
+        
 
 @make_method(mpl.Figure)
 def add_threeaxes(figure,xhist=True,yhist=True,
@@ -643,21 +754,21 @@ def add_threeaxes(figure,xhist=True,yhist=True,
     # ================
     # Input Parsing
     # ================ 
-    if "__iter__" not in dir(shrunk):
+    if not is_arraylike(shrunk):
         shrunk = [shrunk,shrunk]
     elif len(shrunk) == 1:
         shrunk = [shrunk[0],shrunk[0]]
     elif len(shrunk)>2:
         raise ValueError("shrunk cannot have more than 2 entries (x,y)")
 
-    if "__iter__" not in dir(space):
+    if not is_arraylike(space):
         space = [space,space]
     elif len(space) == 1:
         space = [space[0],space[0]]
     elif len(space)>2:
         raise ValueError("space cannot have more than 2 entries (x,y)")
     
-    if "__iter__" not in dir(axspace):
+    if not is_arraylike(axspace):
         axspace = [axspace,axspace]
     elif len(axspace) == 1:
         axspace = [axspace[0],axspace[0]]
@@ -737,7 +848,7 @@ def voronoi_grid(ax,xy):
         simplex = np.asarray(simplex)
         if np.any(simplex < 0):
             i = simplex[simplex >= 0][0]  # finite end Voronoi vertex
-            print i,simplex, simplex[simplex >= 0][0]
+            print(i,simplex, simplex[simplex >= 0][0])
             t = vor.points[pointidx[1]] - vor.points[pointidx[0]]  # tangent
             t /= np.linalg.norm(t)
             n = np.array([-t[1], t[0]])  # normal
